#  Copyright 2022 Diagnostic Image Analysis Group, Radboudumc, Nijmegen, The Netherlands
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from picai_eval.eval import evaluate, evaluate_folder
from picai_eval.metrics import Metrics

print("If you have questions or suggestions, feel free to open an issue " +
      "at https://github.com/DIAGNijmegen/picai_eval\n")


__all__ = [
    # Explicitly expose these functions for easier imports
    "evaluate",
    "evaluate_folder",
    "Metrics",
]
